use handlebars::Handlebars;
use indoc::indoc;
use serde_json::json;

const PKG_NAME: &str = env!("CARGO_PKG_NAME");
const PKG_VERSION: &str = env!("CARGO_PKG_VERSION");

const PAGE_TEMPLATE: &str = indoc! {"
  <html>
    <head>
      <title>{{title}}</title>
    </head>
    <body>
      <style>
        body {
          font-family: 'Helvetica Neue', sans-serif;
          margin: 5em;
        }
        
        h1 {
          font-size: 6em;
          font-weight: bold;
        }
        
        h1 span.subtitle {
          color: #888;
          font-size: 0.45em;
        }
        
        p.server {
          color: #444;
          text-transform: uppercase;
          font-size: 0.8em;
        }

        ul {
          margin-bottom: 4em;
        }
      </style>

      <h1>
        {{title}}
        <br/>
        <span class=\"subtitle\">{{subtitle}}</span>
      </h1>


      {{{content}}}
      
      <p class=\"server\">Served by {{pkg_name}} {{pkg_version}}</p>
    </body>
  </html>
"};

/// Renders an HTML page with the provided title, subtitle, and content.
///
/// The page will include a generate footer with the package name and version.
///
/// NOTE: content is not HTML escaped. It is expected that callers will escape
/// the content before passing it to this function.
pub fn render_server_page(
  title: &str,
  subtitle: &str,
  content: &str,
) -> Result<String, Box<handlebars::TemplateRenderError>> {
  let reg = Handlebars::new();

  reg
    .render_template(
      PAGE_TEMPLATE,
      &json!({
        "title": title,
        "subtitle": subtitle,
        "content": content,
        "pkg_name": PKG_NAME,
        "pkg_version": PKG_VERSION,
      }),
    )
    .map_err(Box::new)
}
